"""
PollenisatorTreeview abstract class
Ttk treeview abstract class to be inherited added functions.
"""
import json
import os
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *

from bson.objectid import ObjectId
from bson.errors import InvalidId
from pollenisatorgui.core.application.dialogs.ChildDialogProgress import ChildDialogProgress
from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.filter import Filter, ParseError
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
import pollenisatorgui.core.components.utils as utils



class PollenisatorTreeview(ttk.Treeview):
    """PollenisatorTreeview class
    Defines common treeview features not implemented by ttk.
    Deletion, expand, collapse, contextualMenu, selection.
    Object stored in a tree view must have a unique iid.
    To make it easier, treeview iid used are their mongo database ID.
    For lists it is given by the view DbToTreeview method.
    """
    dir_path = os.path.dirname(os.path.realpath(__file__))

    def __init__(self, appli, parentFrame):
        """
        Args:
            appli: a reference to the main Application object.
            parentFrame: the parent tkinter window object.
        """
        ttk.Treeview.__init__(self, parentFrame)
        self.appli = appli
        self.parentFrame = parentFrame
        self._detached = []  # Temporary detached objects (filtered objects)
        self._moved = []  # Objects that were moved to be repositioned later
        self._hidden = []  #  Hidden objects reference
        self._hidden_lkp = {}  # Hidden objects lookup table
        self.views = {}  # Dict of views stored in this treeview.
        self.contextualMenu = None
        self.configureTags()
        datamanager = DataManager.getInstance().attach(self)

    def configureTags(self):
        self.icon_oos = utils.loadIcon("ip_oos.png")
        s = Settings()
        oos_color = "gray30" if s.is_dark_mode() else "grey"
        self.tag_configure('OOS', background=oos_color, image=self.icon_oos)
        self.tag_configure('known_command', background="spring green")
        tags = Settings.getTags()
        for tag, color in tags.items():
            if color == "transparent":
                continue
            try:
                self.tag_configure(tag, background=color)
            except tk.TclError:
                #color does not exist
                pass
            
    def resetTags(self, dbId):
        """
        Remove all tags of the node with given id.
        Args:
            dbId: The databaseID of the object to remove tags of
        """
        try:
            # removes all tags, including todo
            self.item(str(dbId), tags=())
        except tk.TclError:
            pass

    def _initContextualsMenus(self):
        """
        Create the contextual menu of variables
        """
        self.contextualMenu = utils.craftMenuWithStyle(self.parentFrame)
        self.contextualMenu.selection = None
        self.contextualMenu.add_command(
            label="Sort children", command=self.sort)
        self.contextualMenu.add_command(
            label="Expand", command=self.expand)
        self.contextualMenu.add_command(
            label="Collapse", command=self.collapse)
        self.contextualMenu.add_command(
            label="Close", command=self.closeMenu)

    def closeMenu(self, _event=None):
        """Does nothing. Used to close the contextual menu."""
        return  # Do nothing.
    
    

    def sort(self, node=None):
        """
        Sort the children node of a treeview node. The sorting key is the node's text.
        Args:
            node: the parent node to sort children of. 
                If none is given, will sort last right clicked node.
                Default is None.
        """
        if node is None:
            try:
                nodeToSort = str(self.contextualMenu.selection)
            except:
                return
        else:
            nodeToSort = node
            
        l = []
        try:
            for k in self.get_children(nodeToSort):
                view_o = self.getViewFromId(str(k))
                if view_o is not None:
                    l.append((k, view_o))
            if l:
                l.sort(key=lambda t: t[1].key() if t[1] is not None else None)
                
                for index, (iid, _) in enumerate(l):
                    # way faster than move is to detach reattach at the end
                    #self.move(iid, nodeToSort, index)
                    self.detach(iid)
                    self.reattach(iid,nodeToSort, "end")
        except tk.TclError: # node given not found
            pass

    def getViewFromId(self, dbId):
        """
        Craft a specific Molde from the Models classes with just a valid Mongo Object Id.

        Args:
            dbId: the database Mongo Id to return a view of.
        """
        try:
            return self.views[dbId]["view"]
        except KeyError:
            return None

    def updateView(self, dbId, model):
        try:
            self.views[dbId]["view"].controller.model = model
        except KeyError:
            pass

    def switchExpandCollapse(self, openAction=True, nodeToExpand=None):
        """
        Expand or collapse all children recursivly of a treeview node.
        Args:
            openAction: Expand if True, Collapse if False.
        """
        if nodeToExpand is None:
            nodeToExpand = str(self.contextualMenu.selection)
        self.item(nodeToExpand, open=openAction)
        children = list(self.get_children(nodeToExpand))
        while len(children) > 0:
            child = children[0]
            children = children + list(self.get_children(child))
            self.item(child, open=openAction)
            del children[0]

    def _getTreeItemState(self, node, toFill):
        """
        Recursive function to get a list of children opened node.
        Args:
            node: the node will want to recursively list opened node of.
            toFill: a list to fill with opened nodes.
        """
        if self.item(node)["open"]:
            toFill.append(str(node))
        children = self.get_children(node)
        for child in children:
            self._getTreeItemState(child, toFill)
        return toFill

    def saveState(self, name):
        """
        Save opened nodes list state to a file.
        file name is given in arguments and stored as an hidden file in Pollenisator/local/states/ folder.
        Args:
            name: the name of this treeview to save.
                  A Dot (".") will be prepended to the name to make the resulting file hidden on linux.
        """
        toFill = []
        ret = self._getTreeItemState('', toFill)
        directory = os.path.join(
            PollenisatorTreeview.dir_path, "../../../local/states/")
        if not os.path.exists(directory):
            os.makedirs(directory)
        path = os.path.join(directory, "."+name)
        with open(path, mode="w") as f:
            f.write(json.dumps(ret))

    def loadState(self, name):
        """
        Load opened nodes list state from a file.
        Restore the state if its exists.
        file name is given in arguments and it must be stored as an hidden file in Pollenisator/local/states/ folder.
        Args:
            name: the name of this treeview to save.
                  The full path to local/states folder and a Dot (".") will be prepended to the name.
        """
        directory = os.path.join(
            PollenisatorTreeview.dir_path, "../../../local/states/")
        if not os.path.exists(directory):
            os.makedirs(directory)
        path = os.path.join(directory, "."+name)
        state = None
        try:
            with open(path, mode="r") as f:
                state = json.loads(f.read())
        except FileNotFoundError:
            state = None
        if state is not None:
            self.restoreTreeItemState(state)

   

    def restoreTreeItemState(self, state):
        """
        Restore the given state.
        Args:
            state: a list of iid to open in the treeview.
        """
        for k in state:
            try:
                self.item(k, open=True)
            except tk.TclError:
                pass

    def open(self, nodeId):
        try:
            self.item(nodeId, open=True)
        except tk.TclError:
            pass

    def expand(self):
        """
        Expand all children recursivly of a treeview node.
        """
        self.switchExpandCollapse(True)

    def collapse(self):
        """
        Collapse all children recursivly of a treeview node.
        """
        self.switchExpandCollapse(False)

    def doPopup(self, event):
        """
        Open the popup contextual menu of the treeview.

        Args:
            event: a ttk Treeview event autofilled. Contains information on what treeview node was clicked.
        """
        # display the popup menu
        self.contextualMenu.selection = self.identify(
            "item", event.x, event.y)
        try:
            self.contextualMenu.tk_popup(event.x_root, event.y_root)
        except Exception as e:
            print(e)
        finally:
            # make sure to release the grab (Tk 8.0a1 only)
            self.contextualMenu.grab_release()
        self.contextualMenu.focus_set()
        self.contextualMenu.bind('<FocusOut>', self.popupFocusOut)

    def popupFocusOut(self, _event=None):
        """Called when the contextual menu loses focus. Closes it.
        Args:
            _event: default to None
        """
        self.contextualMenu.unpost()

    def deleteSelected(self, _event=None):
        """
        Interface to delete a database object from an event.
        Prompt the user a confirmation window.
        Args:
            _event: not used, a ttk Treeview event autofilled. Contains information on what treeview node was clicked.
        """
        n = len(self.selection())
        dialog = ChildDialogQuestion(self.parentFrame,
                                     "DELETE WARNING", "Becareful for you are about to delete "+str(n) + " entries and there is no turning back.", ["Delete", "Cancel"])
        self.wait_window(dialog.app)
        if dialog.rvalue != "Delete":
            return
        if n == 1:
            view = self.getViewFromId(self.selection()[0])
            if view is None:
                return
            view.delete(None, False)
        else:
            toDelete = {}
            for selected in self.selection():
                view = self.getViewFromId(selected)
                if view is not None:
                    viewtype = view.controller.model.coll_name
                    if viewtype not in toDelete:
                        toDelete[viewtype] = []
                    toDelete[viewtype].append(view.controller.getDbId())
            apiclient = APIClient.getInstance()
            apiclient.bulkDelete(toDelete)
                

    def load(self, _event=None):
        """To be overriden
        Args:
            _event: not used, a ttk Treeview event autofilled. Contains information on what treeview node was clicked.
        """
        return

    def onTreeviewSelect(self, _event=None):
        """
        Return ObjectId of selection if it is a valid bson objectid.
        Else return the string of teeview iid.
        Make the viewframe empty.
        Args:
            _event: the treeview node clicked. Not used
        Returns:
            If selection is empty, returns None
            Return ObjectId of selection if it is a valid bson objectid.
            Else return the string of teeview iid.
        """
        selec = self.selection()
        if len(selec) == 0:
            return None
        item = selec[0]
            # the treeview node can either be an object view or a parent node used to store its children and to insert new nodes.
        ret = str(item)
        try:
            # This will raise an exception if the treeview item selected was not a database id.
            # This should only be the case for list of objects ids.
            # An object double click open the list item's type modifying form.
            ret = ObjectId(item)
        except InvalidId:
            pass # str
        if len(self.selection()) == 1:
            for widget in self.appli.viewframe.winfo_children():
                widget.destroy()
        return ret

    def filterTreeview(self, query, settings=None, text_search_allowed=True):
        """
        Deattach objects in the treeview that does not match the query and search settings.
        Args:
            query: filter query string
            settings: a dict of options:
                * "search_exact_match": for exact matching, default to False
                *  "search_show_hidden" : to enable showing hidden objects, default to False
             Default is None.
        Returns:
            True if the filter is done, else if an error occured. Most probably if the query is bad.
        """
        # Reload local settings and prepare search object.
        self.unhideAll()
        searcher = None
        apiclient = APIClient.getInstance()
        if query.strip() != "":
            dialog = ChildDialogProgress(self.parentFrame, "Searching ...",  msg="Querying ...", progress_mode="indeterminate")
            dialog.show()
            try:
                if self.lazyload:
                    if len(query.strip().split(" ")) == 1 and text_search_allowed:
                        textsearch = True
                    else:
                        textsearch = bool(settings.local_settings.get("textsearch", False))
                    searcher = apiclient.searchPentest(query.strip(), textonly=textsearch)
                    dialog.update(msg="Loading in treeview ...")
                    if searcher.get("success", True):
                        self.doFilterTreeview(searcher, True, keep_parents=settings.local_settings.get("keep_parents", True), dialog_progress=dialog)
                    else:
                        tk.messagebox.showerror("Search error", searcher.get("msg", "Unknown error"))
                        return False
                elif len(query.strip().split(" ")) == 1 and text_search_allowed: # 1 word search = text search
                    self.doFilterTreeview(query, False, keep_parents=settings.local_settings.get("keep_parents", True))
                elif settings.local_settings.get("textsearch", False) and text_search_allowed:
                    self.doFilterTreeview(query, False, keep_parents=settings.local_settings.get("keep_parents", True))
                else:
                    searcher = Filter(query, )
                    self.doFilterTreeview(searcher, True, keep_parents=settings.local_settings.get("keep_parents", True))
            except ParseError as e:
                tk.messagebox.showerror("Search error", str(e))
                return False
            finally:
                dialog.destroy()
        return True

    def unfilterAll(self):
        """Reattach all detached objects and reposition them.
        """
        detached = sorted(self._detached, key=lambda x: len(x[0]))
        for detached in self._detached:
            itemId = detached[0]
            parentId = '' if detached[1] is None else detached[1]
            try:
                self.reattach(itemId, parentId, 0)
            except tk.TclError:
                pass
        for moved in self._moved:
            itemId = moved[0]
            parentId = '' if moved[1] is None else moved[1]
            try:
                self.move(itemId, parentId, 0)
            except tk.TclError:
                pass
            view_o = self.getViewFromId(itemId)
            try:
                self.item(itemId, text=str(view_o.controller.getModelRepr()))
            except tk.TclError:
                pass
        for hidden in self._hidden:
            try:
                self.detach(str(hidden[0]))
            except tk.TclError:
                pass
        self._detached = []
        self._moved = []

    def unhide(self, reason):
        hiddens = self._hidden[::-1]
        toDel = []
        for i, hidden in enumerate(hiddens):
            itemId = hidden[0]
            parentId = '' if hidden[1] is None else hidden[1]
            
            if reason in hidden[2]:
                hidden[2].remove(reason)
            
            elif reason == "*":
                hidden[2] = []
            if len(hidden[2]) == 0:
                try:
                    self.reattach(itemId, parentId, 0)
                    toDel.append(i)
                except tk.TclError:
                    pass
        for i in toDel[::-1]:
            if hiddens[i][0] in self._hidden_lkp:
                del self._hidden_lkp[hiddens[i][0]]
            del hiddens[i]
        self._hidden = hiddens[::-1]

    
    def unhideNodeChildren(self, reason, node=None):
        if node is None:
            node = self.selection()[0]
        hiddens = self._hidden[::-1]
        for i, hidden in enumerate(hiddens):
            itemId = hidden[0]
            parentId = '' if hidden[1] is None else hidden[1]
            if str(parentId) != str(node):
                continue
            if reason in hidden[2]:
                hidden[2].remove(reason)
            elif reason == "*":
                hidden[2] = []
            if len(hidden[2]) == 0:
                try:
                    self.reattach(itemId, parentId, 0)
                except tk.TclError:
                    pass

    def unhideAll(self):
        """Reattach all hidden objects but keep in memory that they are hidden.
        """
        hiddens = self._hidden[::-1]
        for hidden in hiddens:
            itemId = hidden[0]
            parentId = '' if hidden[1] is None else hidden[1]
            try:
                self.reattach(itemId, parentId, 0)
            except tk.TclError:
                pass

    

    def doFilterTreeview(self, query, show_hidden=True, keep_parents=True, dialog_progress=None):
        """Apply the query on the treeview.
        Args:
            query: the core.Components.Search object that hold the informations
            show_hidden: will filter the hidden object as well and show them if they match the filter. Default to True.
        """
        # reattach every one
        self.unfilterAll()
        if query is not None:
            if isinstance(query, Filter):
                results_iid = query.getIds(self)
                if len(results_iid) != 0:
                    if show_hidden:
                        self.unhideAll()
                else:
                    tk.messagebox.showerror("No results", "No results found")
                    return
                self._brutSearcher(results_iid, "filter", keep_parents=keep_parents)
            elif isinstance(query, dict):
                total = 0
                results_iid = set()
                datamanager = DataManager.getInstance()
                for value in query.values():
                    total += len(value)
                if total > self.lazylimit:
                    tk.messagebox.showerror("Too many results", f"Too many results found ({total}), please refine your search")
                    return
                try:
                    self.delete("search")
                except tk.TclError:
                    pass
                node = self.insert("", 0, "search", text="Search results")
                for classtype, classitems in query.items():
                    clazz = datamanager.getClass(classtype)
                    if classitems:
                        obj = clazz(classitems[0])
                        view = self.appli.modelToView(clazz.coll_name, obj)
                        if hasattr(view.__class__, "multiAddInTreeview"):
                            # if dialog_progress is not None:
                            #     dialog_progress.update(msg="Adding %s" % clazz.coll_name)
                            view.__class__.multiAddInTreeview(self, self.appli.viewframe, self.appli, [clazz(c) for c in classitems], "search", addChildren=False, detailed=True)
                            results_iid.update(set([str(o.get("_id")) for o in classitems]))
                            continue
                    count = 0
                    size = len(classitems)
                    for classitem in classitems:
                        obj = clazz(classitem)
                        view = None
                        if obj is None:
                            return
                        model = obj
                        # if dialog_progress is not None:
                        #     dialog_progress.update(msg="Adding  %s results in treeview (%s/%s)" % clazz.coll_name, count, size)
                        view = self.appli.modelToView(clazz.coll_name, model)
                        count += 1
                        results_iid.add(str(model.getId()))
                        try:
                            if view is not None:
                                view.addInTreeview(node, addChildren=False)
                                #view.insertReceived()
                        except tk.TclError:
                            pass
                    
                if results_iid:
                    if show_hidden:
                        self.unhideAll()
                if dialog_progress is not None:
                    dialog_progress.update(msg="Filtering treeview ...")
                self._brutSearcher(results_iid, "filter", keep_parents=keep_parents)
            else:
                self._brutSearcher(query, "text", keep_parents=keep_parents)

    # def _brutSearcher(self, query_item, search_type, parentItem='', **kwargs):
    #     stack = [parentItem] if parentItem else list(self.get_children(''))
    #     keep_parents = kwargs.get("keep_parents", True)
    #     while stack:
    #         current_item = stack.pop()
    #         children = list(self.get_children(current_item))

    #         if search_type == "filter":
    #             is_match = current_item in query_item
    #         else:
    #             nodetext = self.item(current_item)["text"]
    #             is_match = str(query_item).lower() in nodetext.lower()

    #         has_child_match = any(child in stack for child in children)  # Check if any children are in the stack

    #         if is_match:
    #             self.item(current_item, open=True)
    #             if not keep_parents:
    #                 self._moved.append([current_item, self.parent(current_item)])
    #                 self.move(current_item, '', 'end')
    #             # Add children to the stack
    #             stack.extend(children)
    #             continue

    #         if keep_parents and has_child_match:
    #             self.item(current_item, open=True)
    #             # Add children to the stack
    #             stack.extend(children)
    #             continue

    #         # If no match and the item has children, add them to the stack
    #         if children:
    #             stack.extend(children)
    #         else:
    #             try:
    #                 self._detached.append([current_item, self.parent(current_item)])
    #                 # Detach the child
    #                 self.detach(current_item)
    #             except tk.TclError:
    #                 pass

    def _brutSearcher(self, query_item, search_type, parentItem='', **kwargs):
        if search_type == "filter":
            is_match = parentItem in query_item
        else:
            nodetext = self.item(parentItem)["text"]
            is_match =  str(query_item).lower() in nodetext.lower()
        keep_parents = kwargs.get("keep_parents", True)
        has_child_match = False
        # Nominal case = empty least
        children = list(self.get_children(parentItem))
        for item_id in children:
            matched = self._brutSearcher(query_item, search_type, item_id, **kwargs)
            if keep_parents and matched:
                has_child_match = True
        if is_match:
            #self.item(parentItem, open=True)
            if not keep_parents:
                self._moved.append([parentItem, self.parent(parentItem)])
                self.move(parentItem, '', 'end')
            return True
        if keep_parents and has_child_match:
            #self.item(parentItem, open=True)
            return True
        try:
            self._detached.append([parentItem, self.parent(parentItem)])
            # Detach the child
            self.detach(parentItem)
        except tk.TclError:
            pass
        return False
    
   


import tkinter as tk
from pollenisatorgui.core.application.dialogs.ChildDialogToast import ChildDialogToast
from pollenisatorgui.core.forms.form import Form

import tkinter.ttk as ttk
from customtkinter import *
import tkinter.messagebox
from pollenisatorgui.core.application.pollenisatorentry import PopoEntry

class FormSearchBar(Form):
    """
    Form field representing a string input.
    """

    def __init__(self, name, searchCallback, panel_to_fill, list_of_forms_to_fill, default="", **kwargs):
        """
        Constructor for a form entry

        Args:
            name: the entry name (id).
            regexValidation: a regex used to check the input in the checkForm function., default is ""
            default: a default value for the Entry, defauult is ""
        """
        super().__init__(name)
        self.searchCallback = searchCallback
        self.panel_to_fill = panel_to_fill
        self.list_of_forms_to_fill = list_of_forms_to_fill
        self.default = default
        self._results = None
        self.kwargs = kwargs
        self.entry = None
        self.result_form = None
        self.options_forms = []
        self.results_forms = {}

    def constructView(self, parent):
        """
        Create the string view inside the parent view given

        Args:
            parent: parent FormPanel.
        """
        self.val = tk.StringVar()
        self.parent = parent
        frame = CTkFrame(parent.panel)
        lbl = CTkLabel(frame, text=self.name+" : ")
        lbl.grid(column=0, row=0)
        self.entry = PopoEntry(frame, textvariable=self.val, width=200)
        self.entry.grid(column=1, row=0)
        self.val.set(self.default)
        
        values = []
        if self.default != "":
            values.append(self.default)
        if self.result_form is None:
            lbl = CTkLabel(frame, text="Search results : ")
            lbl.grid(column=0, row=1)
            self.result_form = CTkComboBox(frame, values=values, width=200, state="readonly")
            self.result_form.grid(column=1, row=1)
        self.result_form.configure( command= self.postSelect)
        self.result_form.bind('<<ComboboxSelected>>', self.postSelect)
        if self.default != "":
            self.result_form.set(self.default)
        
        self.entry.bind('<Key-Return>', self.updateValues)
        if self.getKw("autofocus", False):
            self.entry.focus_set()
        if parent.gridLayout:
            frame.grid(row=self.getKw("row", 0), column=self.getKw("column", 0), **self.kwargs)
        else:
            frame.pack(side=self.getKw("side", "top"), padx=self.getKw("padx", 10), pady=self.getKw("pady", 5), **self.kwargs)
        

   
    def updateValues(self, _event=None):
        options = {}
        for form, option_name in self.options_forms:
            options[option_name] = form.getValue()
        self._results, err_msg = self.searchCallback(self.val.get(), **options)
        if self._results is None:
            tkinter.messagebox.showinfo("SearchBar is not responding", err_msg)
            self.result_form['values'] = [self.val.get()]
            self.result_form.set(self.val.get())
            return
        list_choice = []
        for result in self._results:
            if isinstance(result["TITLE"], str):
                list_choice.append(result["TITLE"])
        self.result_form.configure(values=list_choice)
        if len(list_choice) == 0:
            self.result_form.set("")
            toast = ChildDialogToast(self.parent, "No results found" , 
                                     x=self.result_form.winfo_rootx()+self.result_form.winfo_reqwidth(), y=self.result_form.winfo_rooty(), width=self.result_form.winfo_reqwidth())
            toast.show()
        if len(list_choice) > 0:
            self.result_form.set(list_choice[0])
        if len(list_choice) == 1:
            self.postSelect()

    def fillResults(self, selected):
        for key, value in selected.items():
            if key.lower() in self.results_forms:
                resultform = self.results_forms[key.lower()]
                form = resultform.get("form")
                callback = resultform.get("fill_callback")
                if callback is not None:
                    callback(form, value)
                else:
                    if getattr(form, "addItem", None) is None:
                        form.setValue(value)
                    else:
                        if isinstance(value, dict):
                            form.addItem(**value)
                    
    def postSelect(self, _event=None):
        selected = self.getValue()
        if selected is None:
            return
        if self.results_forms:
            self.fillResults(selected)
            return
        for subform in self.panel_to_fill.subforms:
            if getattr(subform, "subforms", None) is not None: # 1 depth max
                for subform_depth in subform.subforms:
                    if getattr(subform_depth, "subforms", None) is None:
                        if subform_depth.name.lower() in selected.keys():
                            if subform_depth.name.lower() not in [tup[1] for tup in self.options_forms]:
                                if getattr(subform_depth, "addItem", None) is None:
                                    subform_depth.setValue(selected[subform_depth.name.lower()])
                                else:
                                    itemToAdd = selected[subform_depth.name.lower()]
                                    if isinstance(itemToAdd, dict):
                                        subform_depth.addItem(**itemToAdd)
            else:
                if subform.name.lower() not in [tup[1] for tup in self.options_forms]:
                    if subform.name.lower() in selected.keys():
                        if getattr(subform, "addItem", None) is None:
                            subform.setValue(selected[subform.name.lower()])
                        else:
                            itemToAdd = selected[subform.name.lower()]
                            if isinstance(itemToAdd, dict):
                                subform.addItem(**itemToAdd)

    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return the entry value as string.
        """
        title = self.result_form.get()
        if self._results is not None:
            for elem in self._results:
                if elem["TITLE"] == title:
                    return elem
        return None

    def setFocus(self):
        self.entry.focus_set()

    def addOptionForm(self, optionForm, optionName):
        self.options_forms.append((optionForm, optionName))
        optionForm.bind('<Key-Return>', self.updateValues)

    def addResultForm(self, resultForm, resultName, **kwargs):
        self.results_forms[resultName.lower()] = {"form":resultForm, "fill_callback":kwargs.get("fill_callback", None)}
        
    def setResultForm(self, resultForm):
        self.result_form = resultForm


"""User auth infos module to store and use user login informations """
import tkinter as tk
import tkinter.messagebox
import tkinter.ttk as ttk
from customtkinter import *
from bson.objectid import ObjectId
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.modules.module import Module
from pollenisatorgui.core.models.checkitem import CheckItem
from pollenisatorgui.core.application.treeviews.CheatsheetTreeview import CheatsheetTreeview
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.application.scrollableframexplateform import ScrollableFrameXPlateform

class Cheatsheet(Module):
    iconName = "tab_cheatsheet.png"
    tabName = "Cheatsheet"
    coll_name = "cheatsheet"
    order_priority = Module.LOW_PRIORITY
    pentest_types = ["all"]

    def __init__(self, parent, settings, tkApp):
        """
        Constructor
        """
        super().__init__()
        self.dashboardFrame = None
        self.parent = None
        self.infos = {}
        self.treevw = None
        self.tkApp = tkApp
        self.style = None
        self.icons = {}
        self.inited = False
    
    def open(self,view, nbk, treevw):
        apiclient = APIClient.getInstance()
        if self.inited is False:
            self.treevw = treevw
            self.initUI(view)
        if apiclient.getCurrentPentest() is not None:
            self.refreshUI()
        return True

    def refreshUI(self):
        """
        Reload data and display them
        """
        self.loadData()
        self.displayData()

    def loadData(self):
        """
        Fetch data from database
        """
        apiclient = APIClient.getInstance()
        

    def displayData(self):
        """
        Display loaded data in treeviews
        """
        self.treevw.load()

    def initUI(self, parent):
        """
        Initialize Dashboard widgets
        Args:
            parent: its parent widget
        """
        self.inited = True
        self.parent = parent
        self.moduleFrame = CTkFrame(parent)
        #PANED PART
        self.paned = tk.PanedWindow(self.moduleFrame, height=800)
        #RIGHT PANE : Canvas + frame
        
        self.container = CTkFrame(self.paned) # proxy for ScrollableFrame which can't be added to panedWindow
        self.viewframe = ScrollableFrameXPlateform(self.container)
        #LEFT PANE : Treeview
        self.frameTw = CTkFrame(self.paned)
        self.treevw = CheatsheetTreeview(
            self.tkApp, self.frameTw, self.viewframe)
        self.treevw.heading("#0", text="Cheatsheets")
        self.treevw.initUI()
        btn_add_check = CTkButton(self.frameTw, text="Add a check", command=self.createCheck)
        scbVSel = CTkScrollbar(self.frameTw,
                                orientation=tk.VERTICAL,
                                command=self.treevw.yview)
        self.treevw.configure(yscrollcommand=scbVSel.set)
        self.treevw.grid(row=0, column=0, sticky=tk.NSEW)
        scbVSel.grid(row=0, column=1, sticky=tk.NS)
        btn_add_check.grid(row=1, column=0, sticky=tk.S)
        self.paned.add(self.frameTw)
        self.viewframe.pack(fill=tk.BOTH, expand=1)
        self.paned.add(self.container)
        self.paned.pack(fill=tk.BOTH, expand=1)
        self.frameTw.rowconfigure(0, weight=1) # Weight 1 sur un layout grid, sans ça le composant ne changera pas de taille en cas de resize
        self.frameTw.columnconfigure(0, weight=1) # Weight 1 sur un layout grid, sans ça le composant ne changera pas de taille en cas de resize
        self.moduleFrame.pack(padx=10, pady=10, fill=tk.BOTH, expand=True)

    def onCheatsheetDelete(self, event):
        """Callback for a delete key press on a worker.
        Force deletion of worker
        Args:
            event: Auto filled
        """
        apiclient = APIClient.getInstance()
        selected = self.treevw.selection()
        apiclient.bulkDelete({"checkitems":selected}) 

    def createCheck(self, event=None):
        self.treevw.openInsertWindow(CheckItem())

    def update_received(self, dataManager, notif, obj, old_obj):
        if notif["db"] == "pollenisator":
            if self.treevw is not None:
                self.treevw.update_received(dataManager, notif, obj, old_obj)




import multiprocessing
import threading
import socketio
from bson import ObjectId
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.models.tool import Tool
from pollenisatorgui.autoscanworker import executeTool
from pollenisatorgui.core.components.logger_config import logger

class ScanWorker:

    def __init__(self, settings):
        self.settings = settings
        self.sio = socketio.Client()
        self.timer = None
        self.local_scans = dict()

    def wait(self):
        self.sio.wait() 
        
    def connect(self, name):
        self.name = name
        apiclient = APIClient.getInstance()
        self.sio.connect(apiclient.api_url)
        plugins = list(set(self.settings.local_settings.get("my_commands",{}).keys()))
        print("REGISTER "+str(self.name))
        print("supported plugins "+str(plugins))
        self.sio.emit("register", {"name":self.name, "supported_plugins":plugins})
        self.timer = threading.Timer(5.0, self.beacon)
        self.timer.start()
        @self.sio.event
        def executeCommand(data):
            print("GOT EXECUTE "+str(data))
            workerToken = data.get("workerToken")
            pentest = data.get("pentest", "")
            apiclient = APIClient.getInstance()
            apiclient.setConnection(workerToken, pentest)
            toolId = data.get("toolId")
            infos = data.get("infos")
            tool = Tool.fetchObject({"_id":ObjectId(toolId)})
            if tool is None:
                print("Local worker scan was requested but tool not found : "+str(toolId))
                return
            print("Local worker launch task: tool  "+str(toolId))
            self.launchTask(tool, True, self.name, infos=infos)

        @self.sio.event
        def stopCommand(data):
            print("Got stop "+str(data))
            toolId = data.get("tool_iid")
            self.stopTask(toolId)

        @self.sio.event
        def getProgress(data): 
            print("get progress "+str(data))
            toolId = data.get("tool_iid")
            msg = self.getToolProgress(toolId)
            print(msg)
            self.sio.emit("getProgressResult", {"result":msg})

        @self.sio.event
        def deleteWorker(data=None):
            i = 0
            for running in self.local_scans.values():
                running[0].terminate()
                running[0].join()
                break
            self.sio.disconnect()
            if self.timer:
                self.timer.cancel()

    def beacon(self):
        if self.sio.connected:
            self.sio.emit("keepalive", {"name":self.name, "running_tasks":[str(x) for x in self.local_scans]})
            self.timer = threading.Timer(5.0, self.beacon)
            self.timer.start()
            
    def onClosing(self):
        if self.timer:
            self.timer.cancel()
            i = 0
            for running in self.local_scans.values():
                running[0].terminate()
                running[0].join()
                break
            self.sio.disconnect()
            if self.timer:
                self.timer.cancel()

    def launchTask(self, toolModel, checks=True, worker="", infos={}):
        apiclient = APIClient.getInstance()
        launchableToolId = toolModel.getId()
        for item in list(self.local_scans.keys()):
            process_info = self.local_scans.get(item, None)
            if process_info is not None and not process_info[0].is_alive():
                print("Proc finished : "+str(self.local_scans[item]))
                try:
                    del self.local_scans[item]
                except KeyError as e:
                    pass

        #if worker == "" or worker == "localhost" or worker == apiclient.getUser():
        scan = self.local_scans.get(str(launchableToolId), None)
        if scan is not None:
            if scan[0].is_alive() and str(scan[0].pid) != "":
                return
            else:
                del self.local_scans[str(launchableToolId)]
                scan = None

        print("Launch task (start process) , local worker , for tool "+str(toolModel.getId()))
        thread = None
        queue = multiprocessing.Queue()
        queueResponse = multiprocessing.Queue()
        thread = multiprocessing.Process(target=executeTool, args=(queue, queueResponse, apiclient, str(launchableToolId), True, False, (worker == apiclient.getUser()), infos, logger,worker))
        thread.start()
        self.local_scans[str(launchableToolId)] = (thread, queue, queueResponse, toolModel)
        print('Local tool launched '+str(toolModel.getId()))
        # else:
        #     print('laucnh task, send remote tool launch '+str(toolModel.getId()))
        #     apiclient.sendQueueTasks(toolModel.getId())

    def stopTask(self, toolId):
        thread, queue, queueResponse, toolModel = self.local_scans.get(str(toolId), (None, None, None, None))
        if thread is None:
            return False
        try:
            thread.terminate()
        except:
            pass
        try:
            del self.local_scans[str(toolId)]
        except KeyError:
            toolModel.markAsNotDone()
        return True
    
    def getToolProgress(self, toolId):
        thread, queue, queueResponse, toolModel = self.local_scans.get(str(toolId), (None, None,None,None))
        if thread is None or queue is None:
            return ""
        progress = ""
        if not queueResponse.empty():
            progress = queueResponse.get()
            return progress
        if not thread.is_alive():
            return True
        return False
    
    
    def is_local_launched(self, toolId):
        return self.local_scans.get(str(toolId), None) is not None


"""A special form that contains all other forms"""

import tkinter.ttk as ttk
from customtkinter import *
import tkinter as tk
from pollenisatorgui.core.forms.formcombo import FormCombo
from pollenisatorgui.core.forms.formchecklist import FormChecklist
from pollenisatorgui.core.forms.formmarkdown import FormMarkdown
from pollenisatorgui.core.forms.formstr import FormStr
from pollenisatorgui.core.forms.formtext import FormText
#from pollenisatorgui.core.forms.formmarkdown import FormMarkdown
from pollenisatorgui.core.forms.formbutton import FormButton
from pollenisatorgui.core.forms.formdate import FormDate
from pollenisatorgui.core.forms.formlabel import FormLabel
from pollenisatorgui.core.forms.formimage import FormImage
from pollenisatorgui.core.forms.formfile import FormFile
from pollenisatorgui.core.forms.formhidden import FormHidden
from pollenisatorgui.core.forms.formtreevw import FormTreevw
from pollenisatorgui.core.forms.formcheckbox import FormCheckbox
from pollenisatorgui.core.forms.formswitch import FormSwitch
from pollenisatorgui.core.forms.formhelper import FormHelper
from pollenisatorgui.core.forms.formsearchbar import FormSearchBar
from pollenisatorgui.core.forms.formseparator import FormSeparator

from pollenisatorgui.core.forms.form import Form


class FormPanel(Form):
    """
    Form field representing a panel. It is composed of other forms.
    Additional kwargs values:
        grid: set the layout to grid, default to False
    Default setted values:
        width=500
        if pack : padx = 10, pady = 5, side = top, fill = x
        if grid: row = column = 0 sticky = "East"
    """

    def __init__(self, **kwargs):
        """
        Constructor for a panel.
        """
        super().__init__("main panel")
        self.subforms = []
        self.kwargs = kwargs
        self.gridLayout = self.getKw("grid", False)
        self.make_uniform_column = self.getKw("make_uniform_column", None)
        self.save_row_configure = []
        self.save_column_configure = []
        self.panel = None

    def constructView(self, parent):
        """
        Create the panel view by constructing all subforms views inside a tkinter panel the parent view given.
        Args:
            parent: parent view or parent FormPanel.
        """
        if isinstance(parent, FormPanel):  # Panel is a subpanel
            self.panel = CTkFrame(parent.panel,fg_color=self.getKw("fg_color", None), height=self.getKw("height", 0))
        else:
            self.panel = CTkFrame(parent, fg_color=self.getKw("fg_color", None), height=self.getKw("height", 0))
        self.populateView(parent)

    def populateView(self, parent):
        try:
            if self.make_uniform_column is not None:
                self.makeUniformColumn(self.make_uniform_column)
            
            for form in self.subforms:
                form.constructView(self)
            if isinstance(parent, FormPanel):  # Panel is a subpanel
                if parent.gridLayout:
                    # self.panel.grid_rowconfigure(0, weight=1)
                    # self.panel.grid_columnconfigure(0, weight=1)
                    self.panel.grid(column=self.getKw(
                        "column", 0), row=self.getKw("row", 0), sticky=self.getKw("sticky", tk.NSEW), **self.kwargs)
                else:
                    self.panel.pack(fill=self.getKw("fill", "both"), side=self.getKw(
                        "side", "top"), pady=self.getKw("pady", 5), padx=self.getKw("padx", 10), expand=True, **self.kwargs)
            else:  # Master panel, packing
                is_grid = "row" in self.kwargs or "column" in self.kwargs
                if not is_grid:
                    self.panel.pack(fill=self.getKw("fill", "both"), side="top", pady=self.getKw("pady", 5), padx=self.getKw("padx", 30), expand=True)
                else:
                    self.panel.grid(sticky=self.getKw("sticky", tk.NSEW), row=self.getKw("row", 0), column=self.getKw("column", 0), pady=self.getKw("pady", 5), padx=self.getKw("padx", 30))
            for rowconfigured in self.save_row_configure:
                self.panel.grid_rowconfigure(rowconfigured[0], weight=rowconfigured[1])
            for columnconfigured in self.save_column_configure:
                self.panel.grid_columnconfigure(columnconfigured[0], weight=columnconfigured[1])
        except Exception as e:
            raise Exception("Error while populating view of panel named '" + self.name + "' : " + str(e))

    def checkForm(self):
        """
        Check if this form is correctly filled. A panel is correctly filled if all subforms composing it are correctly filled.

        Returns:
            {
                "correct": True if the form is correctly filled, False otherwise.
                "msg": A message indicating what is not correctly filled.
            }
        """
        for form in self.subforms:
            res, msg = form.checkForm()
            if res == False:
                return False, msg
        return True, ""

    def setFocusOn(self, name):
        """Set focus on the given form name
        Args:
            name: the form name to set the focus on
        """
        for form in self.subforms:
            if name == form.name:
                form.setFocus()

    def makeUniformColumn(self, n):
        for i in range(n):
            self.panel.grid_columnconfigure(i, weight=1, uniform="uniform")

    def rowconfigure(self, row, weight):
        self.save_row_configure.append([row, weight]) 

    def columnconfigure(self, column, weight):
        self.save_column_configure.append([column, weight]) 

    def getValue(self):
        """
        Return the form value. Required for a form.
        Returns:
            returns a list of tuple with subform's name and values.
        """
        res = []
        for form in self.subforms:
            if isinstance(form, FormLabel) or isinstance(form, FormButton) or isinstance(form, FormSeparator):
                continue
            val = form.getValue()
            if val is not None:
                if isinstance(form, FormPanel):
                    res += val
                else:
                    res.append((form.name, val))
        return res

    def setValues(self, dict_of_infos):
        """
        Set the form value recursively. 
        Args:
            dict_of_infos: a dictionnary with key = subform's name and value = subform's value.
        """
        for form in self.subforms:
            if isinstance(form, FormLabel) or isinstance(form, FormButton) or isinstance(form, FormSeparator):
                continue
            if form.name in dict_of_infos:
                form.setValue(dict_of_infos[form.name])
            if isinstance(form, FormPanel):
                form.setValues(dict_of_infos)

    def addFormTreevw(self, name, headers,
                      default_values=None, **kwargs):
        """
        Add a form table to this panel.
        Args:
            name: the table desired name
            headers: a list of 2 strings for the table headers
            default_values: a dictionnary with key = column 0 and value = column 1
            kwargs: Keywords for FormTreevw
        """
        ret = FormTreevw(name, headers, default_values, **kwargs)
        self.subforms.append(ret)
        return ret

    def addFormCombo(self, name, choicesList, default=None, **kwargs):
        """
        Add a form combo to this panel.
        Args:
            name: the combobox desired name
            choicesList: a list of options as strings
            default: a string within the choicesList that will be selected by default. optional.
            kwargs: keywords for FormCombo
        """
        formCombo = FormCombo(name, choicesList, default, **kwargs)
        self.subforms.append(formCombo)
        return formCombo

    def addFormChecklist(self, name, choicesList, default=None, values=None, **kwargs):
        """
        Add a form checklist to this panel.
        Args:
            name: the checklist desired name
            choicesList: a list of options as strings
            default: a list of string within the choicesList that will be selected by default
            kwargs: keywords for FormCheckList
        """
        if default is None:
            default = []
        f = FormChecklist(
            name, choicesList, default, values, **kwargs)
        self.subforms.append(f)
        return f

    def addFormCheckbox(self, name, text, default, **kwargs):
        """
        Add a form checkbox to this panel.

        Args:
            name: the checkbox desired name
            text: a label that will be in front of the checkbox
            default: a boolean indicating if the checkbox should be check by default
            kwargs: keywords for FormCheckbox
        """
        f = FormCheckbox(name, text, default, **kwargs)
        self.subforms.append(f)
        return f
    
    def addFormSwitch(self, name, text, default, **kwargs):
        """
        Add a form switch to this panel.

        Args:
            name: the switch desired name
            text: a label that will be in front of the switch
            default: a boolean indicating if the switch should be check by default
            kwargs: keywords for FormSwitch
        """
        f = FormSwitch(name, text, default, **kwargs)
        self.subforms.append(f)
        return f

    def addFormStr(self, name, regexvalidation="", default="", contextualMenu=None, **kwargs):
        """
        Add a form String to this panel.

        Args:
            name: the string var desired name
            regexvalidation: a regex to validate this input
            default: a default value for this input
            width: the width size of the input
            kwargs: keywords for FormStr
        """
        f = FormStr(name, regexvalidation, default, contextualMenu, **kwargs)
        self.subforms.append(f)
        return f

    def addFormSearchBar(self, name, searchCallback, panel_to_fill, default="", **kwargs):
        """
        Add a form String to this panel.

        Args:
            name: the string var desired name
            searchCallback: a callback
            panel_to_fill: panel to search for subforms with namse matching callback ret
            default: a default value for this input
            kwargs: keywords for FormSearchbar
        """
        f = FormSearchBar(
            name, searchCallback, panel_to_fill, None, default, **kwargs)
        self.subforms.append(f)
        return f

    def addFormFile(self, name, regexvalidation="", default="", **kwargs):
        """
        Add a form String to this panel.

        Args:
            name: the string var desired name
            regexvalidation: a regex validating the form
            default: a default value for this input
            kwargs: keywords for FormFile
        """
        f = FormFile(name, regexvalidation, default, **kwargs)
        self.subforms.append(f)
        return f

    def addFormText(self, name, validation="", default="", contextualMenu=None, **kwargs):
        """
        Add a form Text to this panel.

        Args:
            name: the text var desired name
            validation: a regex to validate this input or a callback function
            default: a default value for this input
            kwargs: keywords for FormText
        """
        f = FormText(name, validation, default, contextualMenu, **kwargs)
        self.subforms.append(f)
        return f
    
    def addFormMarkdown(self, name, validation="", default="", contextualMenu=None, **kwargs):
        """
        Add a form Text to this panel.

        Args:
            name: the text var desired name
            validation: a regex to validate this input or a callback function
            default: a default value for this input
            kwargs: keywords for FormText
        """
        f = FormMarkdown(name, validation, default, contextualMenu, **kwargs)
        self.subforms.append(f)
        return f

    def addFormDate(self, name, root, default="", dateformat='%d/%m/%Y %H:%M:%S', **kwargs):
        """
        Add a form Date to this panel.

        Args:
            name: the text var desired name
            default: a default value for this input
            dateformat: a date format to validate this input. Default to dd/mm/YYYY hh:mm:ss.
            kwargs: keywords for FormDate
        """
        f = FormDate(name, root, default, dateformat, **kwargs)
        self.subforms.append(f)
        return f

    def addFormLabel(self, name, text="", **kwargs):
        """
        Add a form Label to this panel.

        Args:
            name: the label desired name
            text: the text printed by the label
            kwargs: keywords for FormLabel
        """
        f = FormLabel(name, text, **kwargs)
        self.subforms.append(f)
        return f

    def addFormImage(self, path, **kwargs):
        """
        Add a form Label to this panel.

        Args:
            path: the image path
            kwargs: keywords for FormImage (same as tk label)
        """
        f = FormImage(path, **kwargs)
        self.subforms.append(f)
        return f

    def addFormHelper(self, helper, **kwargs):
        """
        Add a form Label to this panel.

        Args:
            helper: the text printed by the helper
            kwargs: keywords for FormHelper
        """
        f = FormHelper("", helper, **kwargs)
        self.subforms.append(f)
        return f

    def addFormButton(self, name, callback, **kwargs):
        """
        Add a form Button to this panel.

        Args:
            name: the button desired name and text
            callback: a function that will be called when the button is pressed.
            kwargs: keywords for FormButton
        """
        f = FormButton(name, callback, **kwargs)
        self.subforms.append(f)
        return f
    

    def addFormSeparator(self,  **kwargs):
        """
        Add a form Button to this panel.

        Args:
            callback: a function that will be called when the button is pressed.
            kwargs: keywords for FormButton
        """
        f = FormSeparator(**kwargs)
        self.subforms.append(f)
        return f

    def addFormHidden(self, name, default=""):
        """
        Add a form Hidden to this panel.

        Args:
            name: the hidden value desired name and text
            default: the value to be hidden
        """
        f = FormHidden(name, default)
        self.subforms.append(f)
        return f

    def addFormPanel(self, **kwargs):
        """
        Add a form Pannel to this panel.

        Args:
            kwargs: can indicate grid=True if a grid layout must be set for subelements.
        Returns:
            Return the new panel to add other forms in.
        """
        pan = FormPanel(**kwargs)
        self.subforms.append(pan)
        return pan
    
    def addFormCollapsiblePanel(self, text, **kwargs):
        """
        Add a form Pannel to this panel.

        Args:
            kwargs: can indicate grid=True if a grid layout must be set for subelements.
        Returns:
            Return the new panel to add other forms in.
        """
        from .formcollapsiblepanel import FormCollapsbilePanel
        pan = FormCollapsbilePanel(text, **kwargs)
        self.subforms.append(pan)
        return pan

    def clear(self):
        """
        Empties the panel's subforms.
        """
        del self.subforms
        self.subforms = []

    def redraw(self):
        """
        Empties the panel's subforms.
        """
        if self.panel is None:
            return
        for w in self.panel.winfo_children():
            w.destroy()
        self.populateView(self.panel)



from pollenisatorgui.core.application.dialogs.ChildDialogAskText import ChildDialogAskText
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.settings import Settings
import tkinter as tk
import json
import pandas as pd

def main(apiclient):
    APIClient.setInstance(apiclient)
    apiclient = APIClient.getInstance()
    pentests_list = apiclient.getPentestList()
    pentests = {}
    for pentest in pentests_list:
        pentests[pentest["nom"]] = pentest
    dialog = ChildDialogAskText(None, "Remove unwanted pentests", default="\n".join([pentest["nom"] for pentest in pentests_list]))
    dialog.app.wait_window(dialog.app)
    pentests_name = dialog.rvalue
    defects_stats = []
    for p in pentests_name.split("\n"):
        p = p.strip()
        if p == "":
            continue
        pentest = pentests[p]
        pentest_name = pentest.get("uuid", pentest["nom"])
        pentest_type = apiclient.findInDb(pentest_name, "settings", {"key":"pentest_type"}, False)["value"]
        defects = apiclient.findInDb(pentest_name, "defects", {"ip":""}, True)
        for defect in defects:
            defects_stats.append({"title":defect.get("title",""), "risk":defect.get("risk",""),
                            "date":pentest["creation_date"], "pentest_type":pentest_type})
    f = tk.filedialog.asksaveasfilename(parent=None, defaultextension=".json")
    # asksaveasfile return `None` if dialog closed with "cancel".
    with open(f, "w") as f:
        f.write(json.dumps(defects_stats, indent=4))
    df = pd.DataFrame(defects_stats)
    
    f = tk.filedialog.asksaveasfilename(parent=None, defaultextension=".csv")
    # asksaveasfile return `None` if dialog closed with "cancel".
    df.to_csv(f, index=False)
    return True, f"Dumped"



import json
import multiprocessing
import requests
import os
import io
from datetime import datetime
import sys
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.components.utils import JSONEncoder, JSONDecoder, saveClientConfig
from shutil import copyfile
from jose import jwt, JWTError
from functools import wraps
from bson import ObjectId
import tkinter as tk


dir_path = os.path.dirname(os.path.realpath(__file__))  # fullpath to this file
config_dir = utils.getConfigFolder()
sample_file = os.path.join(utils.getMainDir(), "config/clientSample.cfg")
configClientPath = os.path.join(config_dir, "client.cfg")
if not os.path.isfile(configClientPath):
    if os.path.isfile(sample_file):
        try:
            os.makedirs(config_dir)
        except:
            pass
        copyfile(sample_file, configClientPath)
if os.path.isfile(configClientPath):
    cfg = utils.loadClientConfig()
else:
    print("No client config file found under "+str(configClientPath))
    sys.exit(1)

class ErrorHTTP(Exception):
    def __init__(self, response, *args):
        self.response = response
        self.ret_values = args if args else None

def handle_api_errors(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            res = func(self, *args, **kwargs)
        except requests.exceptions.ProxyError:
            cfg = utils.loadClientConfig()
            cfg["proxies"] = ""
            saveClientConfig(cfg)
            return None
        except requests.exceptions.ConnectionError:
            cfg = utils.loadClientConfig()
            cfg["host"] = ""
            saveClientConfig(cfg)
            return None
        except requests.exceptions.InvalidURL:
            cfg = utils.loadClientConfig()
            cfg["host"] = ""
            saveClientConfig(cfg)
            return None
        except ErrorHTTP as err:
            if err.response.status_code == 401:
                cfg = utils.loadClientConfig()
                cfg["token"] = ""
                saveClientConfig(cfg)
                if "Authorization" in self.headers:
                    del self.headers["Authorization"] 
                err.with_traceback = False
                if self.appli is None:
                    return err.ret_values
                if self.appli.initialized:
                    raise err
                else:
                    return err.ret_values
            elif err.response.status_code == 500:
                raise err
            else:
                if isinstance(err.ret_values, tuple) and len(err.ret_values) == 1:
                    return err.ret_values[0]
                return err.ret_values
        return res
    return wrapper

def call_with_timeout(func, args, kwargs, timeout):
    manager = multiprocessing.Manager()
    return_dict = manager.dict()

    # define a wrapper of `return_dict` to store the result.
    def function(return_dict):
        try:
            return_dict['value'] = func(*args, **kwargs)
        except Exception as e:
            return None

    p = multiprocessing.Process(target=function, args=(return_dict,))
    p.start()

    # Force a max. `timeout` or wait for the process to finish
    p.join(timeout)

    # If thread is still active, it didn't finish: raise TimeoutError
    if p.is_alive():
        p.terminate()
        p.join()
        raise TimeoutError
    else:
        try:
            return return_dict['value']
        except KeyError:
            return None

class APIClient():
    __instances = dict()
    @staticmethod
    def getInstance():
        """ Singleton Static access method.
        """
        pid = os.getpid()  # HACK : One api client per process.
        instance = APIClient.__instances.get(pid, None)
        if instance is None:
            APIClient()
        return APIClient.__instances[pid]
    
    @staticmethod
    def setInstance(apiclient):
        """ Set singleton for current pid"""
        pid = os.getpid()
        APIClient.__instances[pid] = apiclient
    
   

    def searchDefect(self, searchTerms, **kwargs):
        api_url = '{0}report/search'.format(self.api_url_base)
        data = {"type":"defect", "terms":searchTerms, "language":kwargs.get('lang', ""), "perimeter":kwargs.get('perimeter', "")}
        #check_api = kwargs.get('check_api', None)
        #if check_api is not None:
        #    data["check_api"] = check_api
        response = requests.post(api_url, data=json.dumps(data), headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res_obj = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res_obj["answers"], "\n".join(res_obj["errors"])
        else:
            return None, "Unexpected server response "+str(response.status_code)+"\n"+response.text    

    def searchRemark(self, searchTerms, **kwargs):
        api_url = '{0}report/search'.format(self.api_url_base)
        response = requests.post(api_url, data=json.dumps({"type":"remark", "terms":searchTerms, "remark_type":kwargs.get("remark_type", "")}), headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res_obj = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res_obj["answers"], "\n".join(res_obj["errors"])
        elif response.status_code == 204:
            return None, "There is no external knowledge database to query. Check documentation if you have one ready."
        else:
            return None, "Unexpected server response "+str(response.status_code)+"\n"+response.text    

    def __init__(self):
        pid = os.getpid()  # HACK : One mongo per process.
        if APIClient.__instances.get(pid, None) is not None:
            raise Exception("This class is a singleton!")
        self.currentPentest = ""
        self.currentPentestName = ""
        self._observers = []
        self.scope = []
        self.userConnected = None
        self.appli = None
        APIClient.__instances[pid] = self
        self.headers = {'Content-Type': 'application/json'}
        host = cfg.get("host")
        if host is None:
            raise KeyError("config/client.cfg : missing API host value")
        port = cfg.get("port")
        if port is None:
            raise KeyError("config/client.cfg : missing API port value")
        http_proto = "https" if str(cfg.get("https", "True")).title() == "True" else "http"
        self.proxies = cfg.get("proxies", "")
        if isinstance(self.proxies, str):
            self.proxies = {"http": self.proxies, "https": self.proxies}

        self.api_url = http_proto+"://"+host+":"+str(port)+"/"
        self.api_url_base = http_proto+"://"+host+":"+str(port)+"/api/v1/"

    def tryConnection(self, config=None, force=False):
        if config is None:
            config = utils.loadClientConfig()
        if config is None:
            raise FileNotFoundError(str(configClientPath)+" does not exist")
        try:
            is_https = config.get("https", True)
            http_proto = "https" if str(is_https).lower() == "true" or is_https == 1 else "http"
            host = config.get("host")
            port = config.get("port")
            proxies = config.get("proxies", "")
            if isinstance(proxies, str):
                if proxies == "":
                    proxies = {}
                else:
                    proxies = {"http": proxies, "https": proxies}
            self.proxies = proxies
            token = None
            if not force:
                token = config.get("token", None)
            self.api_url = http_proto+"://"+host+":"+str(port)+"/"
            self.api_url_base = http_proto+"://"+host+":"+str(port)+"/api/v1/"
            # requests timeout does not work when the DNS does not respond. So we use a thread to do the request and kill it if it takes too long.
            response = call_with_timeout(requests.get, args=(self.api_url_base,), kwargs={"headers":self.headers, "proxies":self.proxies, "verify":False}, timeout=2)
        except requests.exceptions.RequestException as e:
            return False
        except TimeoutError as e:
            return False
        if response is None:
            return False
        if response.status_code == 200:
            saveClientConfig(config)
            import pollenisatorgui.core.components.settings as settings
            s = settings.Settings()
            if self.api_url not in [x["url"] for x in s.local_settings.get("hosts", [])]:
                s.local_settings["hosts"] = [{"url":self.api_url, "proto":http_proto,"port":port, "host":host}] + s.local_settings.get("hosts", [])
                s.saveLocalSettings()
            self.proxies = proxies
            if token:
                return self.setConnection(token)
        return response.status_code == 200
    
    def tryAuth(self):
        try:
            res = self.setCurrentPentest(self.getCurrentPentest())
        except Exception as e:
            #relog
            return False
        return res

    def reportError(self, err):
        api_url = '{0}issue'.format(self.api_url_base)
        requests.post(api_url, headers=self.headers, data=json.dumps({"error":err}), proxies=self.proxies, verify=False)
        

    def isConnected(self):
        return self.headers.get("Authorization", "") != ""

    
    
    def isAdmin(self):
        return "admin" in self.scope

    def getUser(self):
        return self.userConnected
    
    def disconnect(self):
        self.scope = []
        self.userConnected = None
        self.token = ""
        client_config = utils.loadClientConfig()
        client_config["token"] = self.token
        utils.saveClientConfig(client_config)
        try:
            del self.headers["Authorization"]
        except KeyError:
            pass

    def setConnection(self, token, name=""):
        try:
            jwt_decoded = jwt.decode(token, "", options={"verify_signature":False})
            self.scope = jwt_decoded["scope"]
            self.userConnected = jwt_decoded.get("sub", None)
            self.token = token
            self.headers["Authorization"] = "Bearer "+token
            self.currentPentest = ""
            self.currentPentestName = ""
            for scope in self.scope:
                if scope not in ["pentester", "admin", "user", "owner", "worker"]:
                    self.currentPentest = scope
                    self.currentPentestName = name
            client_config = utils.loadClientConfig()
            client_config["token"] = self.token
            utils.saveClientConfig(client_config)
            
        except JWTError as e:
            return False
        return True

    def getToken(self):
        return self.token

    def login(self, username, passwd):
        api_url = '{0}login'.format(self.api_url_base)
        data = {"username":username, "pwd":passwd}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            token = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return self.setConnection(token)
        return response.status_code == 200

    @handle_api_errors
    def getVersion(self):
        api_url = '{0}version'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            version = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return version
        raise ErrorHTTP(response)
    
    @handle_api_errors
    def setCurrentPentest(self, newCurrentPentest, addDefaultCommands=False):
        if newCurrentPentest.strip() == "":
            self.headers["Authorization"] = ""
            self.scope = []
            self.currentPentest = ""
            self.currentPentestName = ""
            return False
        api_url = '{0}login/{1}'.format(self.api_url_base, newCurrentPentest)
        data = {"addDefaultCommands": addDefaultCommands}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            body = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            token = body["token"]
            pentest_name = body["pentest_name"]
            self.setConnection(token, pentest_name)
            return True
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False)
        
        return False

    def getCurrentPentest(self):
        return self.currentPentest
    
    def getCurrentPentestName(self):
        return self.currentPentestName

    @handle_api_errors
    def unregisterWorker(self, worker_name):
        api_url = '{0}workers/{1}/unregister'.format(self.api_url_base, worker_name)
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def setWorkerInclusion(self, worker_name, setInclusion):
        api_url = '{0}workers/{1}/setInclusion'.format(self.api_url_base, worker_name)
        data = {"db":self.getCurrentPentest(), "setInclusion":setInclusion}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def getRegisteredCommands(self, workerName):
        api_url = '{0}workers/{1}/getRegisteredCommands'.format(self.api_url_base, workerName)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def registeredCommands(self, workerName, commandNames):
        api_url = '{0}workers/{1}/registerCommands'.format(self.api_url_base, workerName)
        response = requests.put(api_url, headers=self.headers, data=json.dumps(commandNames, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def removeCommandFromMyCommands(self, iid):
        api_url = '{0}commands/removeFromMyCommands/{1}'.format(self.api_url_base, iid)
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def addCommandToMyCommands(self, iid):
        api_url = '{0}commands/addToMyCommands/{1}'.format(self.api_url_base, iid)
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def addCommandToWorkerCommands(self, iid):
        api_url = '{0}commands/addToWorkerCommands/{1}'.format(self.api_url_base, iid)
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    def reinitConnection(self):
        self.setCurrentPentest("")

    def attach(self, observer):
        """
        Attach an observer to the database. All attached observers will be notified when a modication is done to a pentest through the methods presented below.

        Args:
            observer: the observer that implements a notify(collection, iid, action) function
        """
        self._observers.append(observer)

    def dettach(self, observer):
        """
        Dettach the given observer from the database.

        Args:
            observer: the observer to detach
        """
        try:
            self._observers.remove(observer)
        except ValueError:
            pass

    @handle_api_errors
    def fetchNotifications(self, pentest, fromTime):
        api_url = '{0}notification/{1}'.format(self.api_url_base, pentest)
        response = requests.get(api_url, headers=self.headers, params={"fromTime":fromTime}, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            notifications = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return notifications
        else:
            return []

    @handle_api_errors
    def getPentestList(self):
        api_url = '{0}pentests'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    
    @handle_api_errors
    def doDeletePentest(self, pentest):
        api_url = '{0}pentest/{1}/delete'.format(self.api_url_base, pentest)
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    @handle_api_errors
    def registerPentest(self, pentest, pentest_type, start_date, end_date, scope, settings, pentesters):
        api_url = '{0}pentest/{1}'.format(self.api_url_base, pentest)
        data = {"pentest_type":str(pentest_type), "start_date":start_date, "end_date":end_date, "scope":scope, "settings":settings, "pentesters":pentesters}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return True, json.loads(response.content.decode('utf-8'))
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, json.loads(response.content.decode('utf-8'), cls=JSONDecoder))
        else:
            return False, json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def find(self, collection, pipeline=None, multi=True):
        try:
            return self.findInDb(self.getCurrentPentest(), collection, pipeline, multi)
        except ErrorHTTP as e:
            raise e


    @handle_api_errors
    def findCommand(self, pipeline=None):
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}commands/find'.format(self.api_url_base)
        data = {"pipeline":(json.dumps(pipeline, cls=JSONEncoder))}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder),  proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def findCheckItem(self, pipeline=None, many=True):
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}cheatsheet/find'.format(self.api_url_base)
        data = {"pipeline":(json.dumps(pipeline, cls=JSONEncoder))}
        data["many"] = many
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder),  proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    
    @handle_api_errors
    def findCheckInstance(self, pipeline=None, many=True):
        apiclient = APIClient.getInstance()
        pentest = apiclient.getCurrentPentest()
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}find/{1}/checkinstances'.format(self.api_url_base, pentest)
        data = {"pipeline":(json.dumps(pipeline, cls=JSONEncoder))}
        data["many"] = many
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder),  proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def updateCheckInstance(self, iid, updatePipeline):
        apiclient = APIClient.getInstance()
        pentest = apiclient.getCurrentPentest()
        api_url = '{0}cheatsheet/{1}/{2}'.format(self.api_url_base, pentest, str(iid))
        response = requests.put(api_url, headers=self.headers, data=json.dumps(updatePipeline, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
        
    @handle_api_errors
    def multiChangeStatus(self, iids, toStatus):
        apiclient = APIClient.getInstance()
        pentest = apiclient.getCurrentPentest()
        api_url = '{0}cheatsheet/{1}/multiChangeOfStatus'.format(self.api_url_base, pentest)
        data = {"iids":iids, "status":toStatus}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def queueCheckInstances(self, iids, priority=0, force=False):
        apiclient = APIClient.getInstance()
        pentest = apiclient.getCurrentPentest()
        api_url = '{0}cheatsheet/{1}/queueCheckInstances'.format(self.api_url_base, pentest)
        data = {"iids":iids, "priority":priority, "force":force}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    
    @handle_api_errors
    def updateCheckItem(self, iid, updatePipeline):
        api_url = '{0}cheatsheet/{1}'.format(self.api_url_base, str(iid))
        response = requests.put(api_url, headers=self.headers, data=json.dumps(updatePipeline, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    
    @handle_api_errors
    def insertCheckItem(self, data):
        api_url = '{0}cheatsheet/'.format(self.api_url_base)
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res["res"], res["iid"]
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
        
    @handle_api_errors
    def apply_check_to_pentest(self, check_iid):
        api_url = '{0}cheatsheet/{1}/{2}'.format(self.api_url_base, self.getCurrentPentest(), check_iid)
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res["res"]
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def deleteCheckInstance(self, iid):
        apiclient = APIClient.getInstance()
        pentest = apiclient.getCurrentPentest()
        api_url = '{0}cheatsheet/{1}/{2}'.format(self.api_url_base, pentest, str(iid))
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def deleteCheckItem(self, iid):
        api_url = '{0}cheatsheet/{1}'.format(self.api_url_base, str(iid))
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def deleteCommand(self, command_iid):
        api_url = '{0}commands/delete/{1}'.format(self.api_url_base, command_iid)
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None


    @handle_api_errors    
    def findInDb(self, pentest, collection, pipeline=None, multi=True, use_cache=True):
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}find/{1}/{2}'.format(self.api_url_base, pentest, collection)
        data = {"pipeline":(json.dumps(pipeline, cls=JSONEncoder)), "many":multi, "use_cache":use_cache}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder),  proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def insert(self, collection, data):
        api_url = '{0}{1}/{2}'.format(self.api_url_base, collection, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder),  proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res["res"], res["iid"]
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None


    @handle_api_errors
    def insertInDb(self, pentest, collection, pipeline=None, parent="", notify=False):
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}insert/{1}/{2}'.format(self.api_url_base, pentest, collection)
        data = {"pipeline":json.dumps(pipeline, cls=JSONEncoder), "parent":parent, "notify":notify}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def update(self, collection, iid, updatePipeline):
        api_url = '{0}{1}/update/{2}/{3}'.format(self.api_url_base, collection, self.getCurrentPentest(), iid)
        response = requests.put(api_url, headers=self.headers,data=json.dumps(updatePipeline, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def updatePentest(self, pentest_name):
        api_url = '{0}pentest/{1}'.format(self.api_url_base, self.getCurrentPentest())
        data = {"pentest_name":pentest_name}
        response = requests.put(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def updatePentestSetting(self, key_value_dict):
        api_url = '{0}settings/{1}'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, data=json.dumps(key_value_dict, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
        
    @handle_api_errors       
    def updateInDb(self, pentest, collection, pipeline, updatePipeline, many=False, notify=False, upsert=False):
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}update/{1}/{2}'.format(self.api_url_base, pentest, collection)
        data = {"pipeline":json.dumps(pipeline, cls=JSONEncoder), "updatePipeline":json.dumps(updatePipeline, cls=JSONEncoder), "many":many, "notify":notify, "upsert":upsert}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def delete(self, collection, iid):
        api_url = '{0}{1}/{2}/{3}'.format(self.api_url_base, collection, self.getCurrentPentest(), iid)
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def bulkDelete(self, dictToDelete):
        api_url = '{0}delete/{1}/bulk'.format(self.api_url_base, self.getCurrentPentest())
        data = dictToDelete
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), verify=False, proxies=self.proxies)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def bulkDeleteCommands(self, dictToDelete, forWorker=False):
        api_url = '{0}commands/delete/bulk'.format(self.api_url_base)
        data = dictToDelete
        if forWorker:
            data["Worker"] = True
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), verify=False, proxies=self.proxies)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    
    @handle_api_errors
    def deleteFromDb(self, pentest, collection, pipeline, many=False, notify=False):
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}delete/{1}/{2}'.format(self.api_url_base, pentest, collection)
        data = {"pipeline":json.dumps(pipeline, cls=JSONEncoder), "many":many, "notify":notify}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def aggregate(self, collection, pipelines=None):
        pipelines = [] if pipelines is None else pipelines
        api_url = '{0}aggregate/{1}/{2}'.format(self.api_url_base, self.getCurrentPentest(), collection)
        data = pipelines
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def count(self, collection, pipeline=None):
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}count/{1}/{2}'.format(self.api_url_base, self.getCurrentPentest(), collection)
        data = {"pipeline":json.dumps(pipeline, cls=JSONEncoder)}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, 0)
        else:
            return 0

    @handle_api_errors
    def getWorkers(self, pipeline=None):
        pipeline = {} if pipeline is None else pipeline
        api_url = '{0}workers'.format(self.api_url_base)
        data = {"pipeline":json.dumps(pipeline, cls=JSONEncoder)}
        response = requests.get(api_url, headers=self.headers, params=data, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code == 204:
            return []
        elif response.status_code >= 400:
            raise ErrorHTTP(response, [])
    
    @handle_api_errors
    def getWorker(self, pipeline=None):
        res = self.getWorkers(pipeline)
        if res is not None:
            if len(res) == 1:
                return res[0]
        return None

    
    @handle_api_errors
    def getSettings(self, pipeline=None):
        if pipeline is None:
            api_url = '{0}settings'.format(self.api_url_base)
            params={}
        else:
            api_url = '{0}settings/search'.format(self.api_url_base)
            params = {"pipeline":json.dumps(pipeline, cls=JSONEncoder)}
        response = requests.get(api_url, headers=self.headers, params=params, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code == 404:
            return []
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def createSetting(self, key, value):
        api_url = '{0}settings/add'.format(self.api_url_base)
        data = {"key":key, "value":json.dumps(value)}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    
    @handle_api_errors
    def updateSetting(self, key, value):
        api_url = '{0}settings/update'.format(self.api_url_base)
        data = {"key":key, "value":json.dumps(value)}
        response = requests.put(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def registerTag(self, pentest, name, color):
        api_url = '{0}settings/registerTag'.format(self.api_url_base)
        data = {"name":name, "color":color, "pentest":pentest}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    
    @handle_api_errors
    def unregisterTag(self,pentest, name):
        api_url = '{0}settings/unregisterTag'.format(self.api_url_base)
        data = {"name":name, "pentest":pentest}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def updateTag(self, name, color, level, isGlobal=False):
        api_url = '{0}settings/updateTag'.format(self.api_url_base)
        if not isGlobal:
            api_url += '/'+self.getCurrentPentest()
        data = {"name":name, "color":color, "level":level, "global":isGlobal}
        response = requests.put(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None


    @handle_api_errors
    def sendStopTask(self, tool_iid, forceReset=False):
        api_url = '{0}tools/{1}/stopTask/{2}'.format(self.api_url_base, self.getCurrentPentest(), tool_iid)
        data = {"forceReset":forceReset}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False,)
        else:
            return False

    @handle_api_errors
    def sendQueueTasks(self, tools_iid):
        if isinstance(tools_iid, str):
            tools_iid = [ObjectId(tools_iid)]
        elif isinstance(tools_iid, ObjectId):
            tools_iid = [tools_iid]
        api_url = '{0}tools/{1}/queueTasks'.format(self.api_url_base, self.getCurrentPentest())
        data = tools_iid
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
        
    @handle_api_errors
    def runTask(self, tool_iid):
        api_url = '{0}tools/{1}/{2}/runTask'.format(self.api_url_base, self.getCurrentPentest(), tool_iid)
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

        
    @handle_api_errors
    def sendRemoveTasks(self, tools_iid):
        if isinstance(tools_iid, str):
            tools_iid = [ObjectId(tools_iid)]
        elif isinstance(tools_iid, ObjectId):
            tools_iid = [tools_iid]
        api_url = '{0}tools/{1}/removeTasks'.format(self.api_url_base, self.getCurrentPentest())
        data = tools_iid
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
        
    @handle_api_errors  
    def getQueue(self):
        api_url = '{0}tools/{1}/getQueue'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
        
    @handle_api_errors
    def clear_queue(self):
        api_url = '{0}tools/{1}/clearTasks'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None


    @handle_api_errors
    def addCustomTool(self, port_iid, command_iid):
        api_url = '{0}ports/{1}/{2}/addCustomTool/'.format(self.api_url_base, self.getCurrentPentest(), port_iid)
        response = requests.post(api_url, headers=self.headers, data=json.dumps({"command_iid":command_iid}, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def putProof(self, defect_iid, local_path):
        api_url = '{0}files/{1}/upload/proof/{2}'.format(self.api_url_base, self.getCurrentPentest(), defect_iid)
        if os.path.exists(local_path) and os.path.isfile(local_path):
            with open(local_path,mode='rb') as f:
                h = self.headers.copy()
                h.pop("Content-Type", None)
                response = requests.post(api_url, files={"upfile": (os.path.basename(local_path) ,f)}, headers=h, proxies=self.proxies, verify=False)
                if response.status_code == 200:
                    return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
                elif response.status_code >= 400:
                    raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def listProofs(self, defect_iid):
        api_url = '{0}files/{1}/download/proof/{2}'.format(self.api_url_base, self.getCurrentPentest(), defect_iid)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, [])
        return []

    @handle_api_errors
    def getResult(self, tool_iid, local_path):
        resultfile = self.getFileName("result", tool_iid)
        if resultfile is not None:
            try:
                return self._get("result", tool_iid, resultfile, local_path)
            except ErrorHTTP as e:
                raise e
        return None

    @handle_api_errors
    def getProof(self, defect_iid, filename, local_dir):
        try:
            return self._get("proof", defect_iid, filename, local_dir)
        except ErrorHTTP as e:
            raise e
        
    
    @handle_api_errors
    def getFileName(self, filetype, attached_iid):
        """Retrieve the list of filenames attached to a toolid
        Args:
            filetype: 'result' or 'proof' 
            attached_iid: tool or defect iid depending on filetype
        Returns : filename: remote file file name
        """
        api_url = '{0}files/{1}/download/{2}/{3}'.format(self.api_url_base, self.getCurrentPentest(), filetype, str(attached_iid))
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            ret = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            if isinstance(ret, list) and ret:
                return ret[0]
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def _get(self, filetype, attached_iid, filename, local_path):
        """Download file affiliated with given iid and place it at given path
        Args:
            filetype: 'result' or 'proof' 
            attached_iid: tool or defect iid depending on filetype
            filename: remote file file name
            local_path: local file path
        """
        api_url = '{0}files/{1}/download/{2}/{3}/{4}'.format(self.api_url_base, self.getCurrentPentest(), filetype, attached_iid, filename)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            if not os.path.exists(local_path):
                os.makedirs(local_path)
            local_path = os.path.join(local_path, filename)
            with open(local_path, mode='wb') as f:
                f.write(response.content)
            return local_path
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def rmProof(self, defect_iid, filename):
        """Remove file affiliated with given iid 
        """
        api_url = '{0}files/{1}/{2}/{3}'.format(self.api_url_base, self.getCurrentPentest(), defect_iid, filename)
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def getDesiredOutputForPlugin(self, cmdline, plugin):
        """Complete information for a command line. Plugin can be 'auto-detect', a marker for |outputDir| is to be replaced
        """
        api_url = '{0}tools/getDesiredOutputForPlugin'.format(self.api_url_base)
        response = requests.post(api_url, headers=self.headers, data=json.dumps({"plugin":plugin, "cmdline":cmdline}), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            data = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return True, data
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.content.decode('utf-8'))
        return False, response.content.decode('utf-8')
    
    @handle_api_errors
    def getCommandLine(self, toolId, commandline_options=""):
        """Get full command line from toolid and choosen parser, a marker for |outputDir| is to be replaced
        """
        api_url = '{0}tools/{1}/craftCommandLine/{2}'.format(self.api_url_base, self.getCurrentPentest(), toolId)
        response = requests.get(api_url, headers=self.headers, params={"commandline_options":commandline_options}, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            data = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return True, data
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.content.decode('utf-8'))
        return False, response.content.decode('utf-8')
    
    @handle_api_errors
    def importToolResult(self, tool_iid, parser, local_path):
        api_url = '{0}tools/{1}/importResult/{2}'.format(self.api_url_base, self.getCurrentPentest(), tool_iid)
        if not os.path.isfile(local_path):
            return "Failure to open provided file"
        with io.open(local_path, mode='rb') as f:
            h = self.headers.copy()
            h.pop("Content-Type", None)
            response = requests.post(api_url, files={"upfile": (os.path.basename(local_path) ,f)}, data={"plugin":parser}, headers=h, proxies=self.proxies, verify=False)
            if response.status_code == 200:
                return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            elif response.status_code >= 400:
                raise ErrorHTTP(response, response.content.decode('utf-8'))
            return response.content.decode('utf-8')

    @handle_api_errors
    def setToolStatus(self, tool_iid, newStatus, arg=""):
        api_url = '{0}tools/{1}/{2}/changeStatus'.format(self.api_url_base, self.getCurrentPentest(), tool_iid)
        response = requests.post(api_url, headers=self.headers, data=json.dumps({"newStatus":newStatus, "arg":arg}), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def importExistingResultFile(self, filepath, plugin, default_target={}, command_used=""):
        api_url = '{0}files/{1}/import'.format(self.api_url_base, self.getCurrentPentest())
        with io.open(filepath, mode='rb') as f:
            h = self.headers.copy()
            h.pop("Content-Type", None)
            response = requests.post(api_url, headers=h, files={"upfile": (os.path.basename(filepath) ,f)}, data={"plugin":plugin, "default_target":json.dumps(default_target), "cmdline":command_used}, proxies=self.proxies, verify=False)
            if response.status_code == 200:
                return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            elif response.status_code >= 400:
                raise ErrorHTTP(response, json.loads(response.content.decode('utf-8'), cls=JSONDecoder))
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def fetchWorkerInstruction(self, worker_name):
        api_url = '{0}workers/{1}/instructions'.format(self.api_url_base, worker_name)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            data = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return data
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def deleteWorker(self, worker_name):
        api_url = '{0}workers/{1}'.format(self.api_url_base, worker_name)
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False, timeout=2)
        if response.status_code == 200:
            data = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return data
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None
    
    @handle_api_errors
    def deleteWorkerInstruction(self, worker_name, instruction_iid):
        api_url = '{0}workers/{1}/instructions/{2}'.format(self.api_url_base, worker_name, instruction_iid)
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            data = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return data
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def sendStartAutoScan(self, command_iids=None, autoqueue=False):
        if command_iids is None:
            command_iids = []
        api_url = '{0}autoscan/{1}/start'.format(self.api_url_base, self.getCurrentPentest())
        data = {"command_iids":command_iids, "autoqueue":bool(autoqueue)}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None
    
    @handle_api_errors
    def sendStopAutoScan(self):
        api_url = '{0}autoscan/{1}/stop'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def getAutoScanStatus(self):
        api_url = '{0}autoscan/{1}/status'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None        

    @handle_api_errors
    def exportCommands(self, parent):
        api_url = '{0}exportCommands'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            filename = "commands.json" 
            dir_path = os.path.dirname(os.path.realpath(__file__))
            out_path = os.path.normpath(os.path.join(
                dir_path, "../../exports/"))
            f = tk.filedialog.asksaveasfilename(parent=parent, defaultextension=".json", initialdir=out_path, initialfile=filename)
            if f is None or len(f) == 0:  # asksaveasfile return `None` if dialog closed with "cancel".
                return
            filename = str(f)
            with open(filename, mode='wb') as f:
                f.write(response.content)
                return True, filename
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        return False, response.text  

    @handle_api_errors
    def exportCheatsheet(self, parent):
        api_url = '{0}exportCheatsheet'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            filename = "cheatsheet.json" 
            dir_path = os.path.dirname(os.path.realpath(__file__))
            out_path = os.path.normpath(os.path.join(
                dir_path, "../../exports/"))
            f = tk.filedialog.asksaveasfilename(parent=parent, defaultextension=".json", initialdir=out_path, initialfile=filename)
            if f is None or len(f) == 0:  # asksaveasfile return `None` if dialog closed with "cancel".
                return
            filename = str(f)
            with open(filename, mode='wb') as f:
                f.write(response.content)
                return True, filename
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        return False, response.text    
    
    @handle_api_errors
    def exportDefectTemplates(self, parent):
        api_url = '{0}report/DefectTemplates/export'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            filename = "defect_templates.json" 
            dir_path = os.path.dirname(os.path.realpath(__file__))
            out_path = os.path.normpath(os.path.join(
                dir_path, "../../exports/"))
            f = tk.filedialog.asksaveasfilename(parent=parent, defaultextension=".json", initialdir=out_path, initialfile=filename)
            if f is None or len(f) == 0:  # asksaveasfile return `None` if dialog closed with "cancel".
                return 
            filename = str(f)
            with open(filename, mode='wb') as f:
                f.write(response.content)
                return True, filename
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        return False, response.text    

    @handle_api_errors
    def dumpDb(self, parent, pentest, collection=""):
        api_url = '{0}dumpDb/{1}'.format(self.api_url_base, pentest)
        response = requests.get(api_url, headers=self.headers, params={"collection":collection}, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            dir_path = os.path.dirname(os.path.realpath(__file__))
            out_path = os.path.join(
                dir_path, "../../exports/")
            f = tk.filedialog.asksaveasfilename(parent=parent, defaultextension=".gz", initialdir=out_path, initialfile=(pentest if collection == "" else pentest+"_"+collection)+".gz")
            if f is None or len(f) == 0:  # asksaveasfile return `None` if dialog closed with "cancel".
                return None, None
            filename = str(f)          
            with open(filename, mode='wb') as f:
                f.write(response.content)
                return True, filename
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        return False, response.text  

    @handle_api_errors
    def importDb(self, filename, original_db_name):
        api_url = '{0}importDb'.format(self.api_url_base)
        with io.open(filename, mode='rb') as f:
            h = self.headers.copy()
            h.pop("Content-Type", None)
            response = requests.post(api_url, headers=h, files={"upfile": (os.path.basename(filename) ,f,'application/gzip')}, params={"orig_name":original_db_name}, proxies=self.proxies, verify=False)
            if response.status_code >= 400:
                raise ErrorHTTP(response, False)
            return response.status_code == 200
        return False
    
    @handle_api_errors
    def importCommands(self, filename, forWorker=False):
        api_url = '{0}importCommands'.format(self.api_url_base)
        if forWorker:
            api_url += "/worker"
        with io.open(filename, mode='rb') as f:
            h = self.headers.copy()
            h.pop("Content-Type", None)
            response = requests.post(api_url, headers=h, files={"upfile": (os.path.basename(filename) ,f, 'application/json')}, proxies=self.proxies, verify=False)
            if response.status_code >= 400:
                raise ErrorHTTP(response, False)
            return response.status_code == 200

    @handle_api_errors
    def importCheatsheet(self, filename):
        api_url = '{0}importCheatsheet'.format(self.api_url_base)
        with io.open(filename, mode='rb') as f:
            h = self.headers.copy()
            h.pop("Content-Type", None)
            response = requests.post(api_url, headers=h, files={"upfile": (os.path.basename(filename) ,f, 'application/json')}, proxies=self.proxies, verify=False)
            if response.status_code >= 400:
                raise ErrorHTTP(response, False)
            return response.status_code == 200

    @handle_api_errors
    def importDefectTemplates(self, filename):
        api_url = '{0}report/DefectTemplates/import'.format(self.api_url_base)
        with io.open(filename, mode='rb') as f:
            h = self.headers.copy()
            h.pop("Content-Type", None)
            response = requests.post(api_url, headers=h, files={"upfile": (os.path.basename(filename) ,f, 'application/json')}, proxies=self.proxies, verify=False)
            if response.status_code >= 400:
                raise ErrorHTTP(response, False)
            return response.status_code == 200
        
    @handle_api_errors
    def findDefectTemplateById(self, iid):
        api_url = '{0}report/DefectTemplates/find'.format(self.api_url_base)
        response = requests.post(api_url, headers=self.headers, data=json.dumps({"_id":ObjectId(iid)}, cls=JSONEncoder),  proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def insertAsTemplate(self, data):
        api_url = '{0}report/DefectTemplates/insert'.format(self.api_url_base)
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder),  proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res["res"], res["iid"]
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None

    @handle_api_errors
    def updateDefectTemplate(self, iid, updatePipeline):
        api_url = '{0}report/DefectTemplates/update/{1}'.format(self.api_url_base, str(iid))
        response = requests.put(api_url, headers=self.headers, data=json.dumps(updatePipeline, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        else:
            return None
    
    @handle_api_errors
    def deleteDefectTemplate(self, iid):
        api_url = '{0}report/DefectTemplates/delete/{1}'.format(self.api_url_base, str(iid))
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def copyDb(self, fromDb, toDb=""):
        api_url = '{0}copyDb'.format(self.api_url_base)
        data = {"fromDb":self.getCurrentPentest(), "toDb":toDb}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None
 
    @handle_api_errors
    def generateReport(self, templateName, clientName, contractName, mainRedac, lang, additional_context):
        api_url = '{0}report/{1}/generate'.format(self.api_url_base, self.getCurrentPentest())
        data = {"templateName":templateName,"mainRedactor":mainRedac, "lang":lang, "additional_context":additional_context}
        response = requests.post(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            timestr = datetime.now().strftime("%Y%m")
            ext = os.path.splitext(templateName)[-1]
            basename = clientName.strip()+" - "+contractName.strip()
            out_name = str(timestr)+" - "+basename
            out_path = os.path.join(dir_path, "../../exports/")
            f = tk.filedialog.asksaveasfilename(defaultextension=ext, initialdir=out_path, initialfile=out_name+ext)
            if f is None or len(f) == 0:  # asksaveasfile return `None` if dialog closed with "cancel". and empty tuple if close by cross
                return False, "No report created"
            filename = str(f)
            with open(filename, mode='wb') as f:
                f.write(response.content)
                return True, os.path.normpath(filename)
        elif response.status_code >= 400:
            return False, response.content.decode("utf-8")
        return False, response.content.decode("utf-8")

    @handle_api_errors
    def getTemplateList(self, lang):
        api_url = '{0}report/{1}/templates'.format(self.api_url_base, lang)
        try:
            response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        except ConnectionError as e:
            raise e
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, [])
        return []

    @handle_api_errors
    def getLangList(self):
        api_url = '{0}report/langs'.format(self.api_url_base)
        try:
            response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        except ConnectionError as e:
            raise e
        if response.status_code == 200:
            return json.loads(response.content.decode("utf-8"), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, ["en"])
        return ["en"]
    
    @handle_api_errors
    def downloadTemplate(self, parent, lang, templateName):
        api_url = '{0}report/{1}/templates/download'.format(self.api_url_base, lang)
        response = requests.get(api_url, headers=self.headers, params={"templateName":templateName},proxies=self.proxies, verify=False)
        if response.status_code == 200:
            out_path = os.path.join(dir_path, "../../exports/")
            f = tk.filedialog.asksaveasfilename(parent=parent, initialdir=out_path, initialfile=templateName)
            if f is None or len(f) == 0:  # asksaveasfile return `None` if dialog closed with "cancel".
                return
            filename = str(f)
            with open(filename, mode='wb') as f:
                f.write(response.content)
                return os.path.normpath(filename)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def getDefectTable(self):
        api_url = '{0}report/{1}/defects'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, [])
        return []

    @handle_api_errors
    def moveDefect(self, defect_id, target_id):
        api_url = '{0}report/{1}/defects/move/{2}/{3}'.format(self.api_url_base, self.getCurrentPentest(), defect_id, target_id)
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, response.content.decode("utf-8"))
        return response.content.decode("utf-8")

    @handle_api_errors
    def getUsers(self):
        api_url = '{0}admin/listUsers'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, [])
        return []

    @handle_api_errors
    def searchUsers(self, username):
        api_url = '{0}user/searchUsers/{1}'.format(self.api_url_base, username)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, [])
        return []

    @handle_api_errors
    def registerUser(self, username, password, name, surname, email):
        api_url = '{0}user/register'.format(self.api_url_base)
        response = requests.post(api_url, headers=self.headers, data=json.dumps({"username":username, "pwd":password, "name":name, "surname":surname,"email":email}), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def updateUserInfos(self, username, name, surname, email):
        api_url = '{0}user/updateUserInfos'.format(self.api_url_base)
        response = requests.post(api_url, headers=self.headers, data=json.dumps({"username":username, "name":name, "surname":surname,"email":email}), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def deleteUser(self, username):
        api_url = '{0}user/delete/{1}'.format(self.api_url_base, username)
        response = requests.delete(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response)
        return None

    @handle_api_errors
    def changeUserPassword(self, oldPwd, newPwd):
        api_url = '{0}user/changePassword'.format(self.api_url_base)
        response = requests.post(api_url, headers=self.headers, data=json.dumps({"oldPwd":oldPwd, "newPwd":newPwd}), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return ""
        elif response.status_code >= 400:
            raise ErrorHTTP(response, json.loads(response.content.decode('utf-8'), cls=JSONDecoder))
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def resetPassword(self, username, newPwd):
        api_url = '{0}admin/resetPassword'.format(self.api_url_base)
        response = requests.post(api_url, headers=self.headers, data=json.dumps({"username":username, "newPwd":newPwd}), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return ""
        elif response.status_code >= 400:
            raise ErrorHTTP(response, json.loads(response.content.decode('utf-8'), cls=JSONDecoder))
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def getPlugins(self):
        api_url = '{0}tools/plugins'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, [])
        return []

    @handle_api_errors
    def getDockerForPentest(self, pentest):
        api_url = '{0}workers/start/{1}'.format(self.api_url_base, pentest)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code == 403:
            tk.messagebox.showerror("Docker error", "Docker could not start, check server installation of docker")
            return None
        elif response.status_code >= 400:
            raise ErrorHTTP(response, [])
        return None

    @handle_api_errors
    def getComputerUsers(self, computer_iid):
        api_url = '{0}ActiveDirectory/computers/{1}/{2}/getUsers'.format(self.api_url_base, self.getCurrentPentest(), computer_iid)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code == 403:
            tk.messagebox.showerror("Docker error", "Docker could not start, check server installation of docker")
            return None
        elif response.status_code >= 400:
            raise ErrorHTTP(response, {})
        return None

    @handle_api_errors
    def linkAuth(self, auth_iid, object_iid):
        api_url = '{0}auth/{1}/{2}/link/{3}'.format(self.api_url_base, self.getCurrentPentest(), auth_iid, object_iid)
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return response.content.decode('utf-8')
        elif response.status_code >= 400:
            raise ErrorHTTP(response, "Error : "+str(response.content.decode('utf-8')))
        return "Undefined return status code"

    @handle_api_errors
    def getModuleInfo(self, module):
        api_url = '{0}{1}/getModuleInfo'.format(self.api_url_base, module)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, "Error : "+str(response.content.decode('utf-8')))
        return "Undefined return status code"

    @handle_api_errors
    def addRangeMatchingIps(self):
        api_url = '{0}NetworkDiscovery/{1}/addRangeMatchingIps'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return ""
        elif response.status_code >= 400:
            raise ErrorHTTP(response, json.loads(response.content.decode('utf-8'), cls=JSONDecoder))
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)


    @handle_api_errors
    def addRangeCloseToOthers(self):
        api_url = '{0}NetworkDiscovery/{1}/addRangeCloseToOthers'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return ""
        elif response.status_code >= 400:
            raise ErrorHTTP(response, json.loads(response.content.decode('utf-8'), cls=JSONDecoder))
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
    
    @handle_api_errors
    def addCommonRanges(self):
        api_url = '{0}NetworkDiscovery/{1}/addCommonRanges'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return ""
        elif response.status_code >= 400:
            raise ErrorHTTP(response, json.loads(response.content.decode('utf-8'), cls=JSONDecoder))
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def addAllLANRanges(self):
        api_url = '{0}NetworkDiscovery/{1}/addAllLANRanges'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return ""
        elif response.status_code >= 400:
            raise ErrorHTTP(response, json.loads(response.content.decode('utf-8'), cls=JSONDecoder))
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def getCheckInstanceRepr(self, checkinstance_iids):
        api_url = '{0}cheatsheet/{1}/getTargetRepr'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, data=json.dumps(checkinstance_iids, cls=JSONEncoder), proxies=self.proxies ,verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, {},)
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def getCheckInstanceInfo(self, checkinstance_iid):
        api_url = '{0}cheatsheet/{1}/{2}'.format(self.api_url_base, self.getCurrentPentest(), checkinstance_iid)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies ,verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, {},)
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def getDefectTargetRepr(self, defect_iids):
        api_url = '{0}defects/{1}/getTargetRepr'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.post(api_url, headers=self.headers, data=json.dumps(defect_iids, cls=JSONEncoder), proxies=self.proxies ,verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, {},)
        return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)

    @handle_api_errors
    def getToolProgress(self, tool_iid):
        api_url = '{0}tools/{1}/getProgress/{2}'.format(self.api_url_base, self.getCurrentPentest(), tool_iid)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return None

    @handle_api_errors
    def getToolDetailedString(self, tool_iid):
        api_url = '{0}tools/{1}/{2}/getDetailedString'.format(self.api_url_base, self.getCurrentPentest(), tool_iid)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return "Error reaching API"

    @handle_api_errors
    def getTriggerLevels(self):
        api_url = '{0}getTriggerLevels'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return []
        
    @handle_api_errors
    def getPluginTags(self):
        api_url = '{0}tags/getPluginTags'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            tags_per_plugin = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            tags = []
            for plugin in tags_per_plugin:
                tags += tags_per_plugin[plugin]
            return tags
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return []

    @handle_api_errors
    def getCommandVariables(self):
        api_url = '{0}getCommandVariables'.format(self.api_url_base)
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return []
            
    @handle_api_errors
    def getGeneralInformation(self):
        api_url = '{0}pentest/{1}'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.get(api_url, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return []
        
    @handle_api_errors
    def addTag(self, item_id, item_type, tag, overrideGroup=True):
        api_url = '{0}tags/{1}/addTag/{2}'.format(self.api_url_base, self.getCurrentPentest(), item_id)
        data = {"item_type":item_type,"tag":tag, "overrideGroup":overrideGroup}
        response = requests.put(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return None
    
    @handle_api_errors
    def delTag(self, item_id, item_type, tag):
        api_url = '{0}tags/{1}/delTag/{2}'.format(self.api_url_base, self.getCurrentPentest(), item_id)
        data = {"item_type":item_type,"tag":tag}
        response = requests.put(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return None
        
    @handle_api_errors
    def setTags(self, item_id, item_type, tags):
        api_url = '{0}tags/{1}/setTags/{2}'.format(self.api_url_base, self.getCurrentPentest(), item_id)
        data = {"item_type":item_type,"tags":tags}
        response = requests.put(api_url, headers=self.headers, data=json.dumps(data, cls=JSONEncoder), proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
            return None
        
    @handle_api_errors
    def searchTaggedBy(self, tag_name):
        api_url = '{0}tags/{1}/getTaggedBy'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.get(api_url, params={"tag_name":tag_name}, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
        elif response.status_code >= 400:
            raise ErrorHTTP(response, False, response.text)
        else:
             return {"success":False, "msg":"Unexpected server response "+str(response.status_code)+"\n"+response.text}
        
    @handle_api_errors
    def searchPentest(self, stringQuery, textonly=False):
        api_url = '{0}search/{1}'.format(self.api_url_base, self.getCurrentPentest())
        response = requests.get(api_url, params={"s":stringQuery,"textonly":textonly}, headers=self.headers, proxies=self.proxies, verify=False)
        if response.status_code == 200:
            res_obj = json.loads(response.content.decode('utf-8'), cls=JSONDecoder)
            return res_obj
        else:
            return {"success":False, "msg":"Unexpected server response "+str(response.status_code)+"\n"+response.text}


from bson import ObjectId
from pollenisatorgui.core.components.apiclient import APIClient
import pollenisatorgui.core.components.utils as utils
import tempfile
import os
from pollenisatorgui.core.application.dialogs.ChildDialogAskText import ChildDialogAskText

import shutil

def main(apiclient, appli, **kwargs):
    users_to_test = [
    "test",
    "admin",
    "backup",
    "administrateur",
    "Administrateur",
    "administrator",
    "testad",
    "ad",
    "stage",
    "audit",
    "pentest",
    "pwn",
    "super",
    "preprod",
    "prod",
    "demo",
    "visio",
    "livesync",
    "vagrant",
    "guest",
    "glpiadmin",
    "adminglpi",
]
    cme_path = utils.which_expand_alias("cme")
    if not cme_path:
        return False, "binary 'cme' is not in the PATH."
    APIClient.setInstance(apiclient)
    domain = ""
    if kwargs.get("target_type") == "computer":
        computer_info =  apiclient.find("computers", {"type":"computer", "_id":ObjectId(kwargs.get("target_iid", ""))}, False)
        if computer_info is not None:
            domain = computer_info.get("domain", "")
    if domain == "":
        dialog = ChildDialogAskText(None, "Enter domain name", multiline=False)
        dialog.app.wait_window(dialog.app)
        domain = dialog.rvalue
    if domain == "" or domain is None:
        return False, "No domain given"
    dialog = ChildDialogAskText(None, "Basic users to try", "\n".join(users_to_test))
    dialog.app.wait_window(dialog.app)
    if dialog.rvalue is None:
        return False, "Canceled"
    users_to_test = [r.strip() for r in dialog.rvalue.split("\n")]
    dc = None
    exec = 0
    dc_info = apiclient.find("computers", {"type":"computer", "domain":domain, "infos.is_dc":True}, False)
    if dc_info is None:
        dialog = ChildDialogAskText(None, "DC not known, give me IP if you know it", multiline=False)
        dialog.app.wait_window(dialog.app)
        dc = dialog.rvalue
    else:
        dc = dc_info.get("ip")
    if dc is None or dc == "":
        return False, "DC not known"
    temp_folder = tempfile.gettempdir() 
    file_name = os.path.join(temp_folder, "users_"+str(domain)+".txt")
    users = "\n".join(users_to_test)
    if users.strip() != "":
        with open(file_name, "w") as f:
            f.write(users+"\n")
        exec += 1
        appli.launch_in_terminal(kwargs.get("default_target",None), "CME try common users", f"{cme_path} smb {dc} -u {file_name} -p {file_name} -d {domain} --no-bruteforce --continue-on-success"),
    return True, f"Launched {exec} cmes"



"""Describe tkinter Text with default common args and an attached scrollbar"""

import tkinter as tk
from customtkinter import *
from pollenisatorgui.core.application.mdframe import TkintermdFrame
from tkinter.constants import *
from pollenisatorgui.core.forms.form import Form
import pollenisatorgui.core.components.utils as utils


class FormMarkdown(Form):
    """
    Form field representing a multi-lined input.
    Default setted values:
        width=20, height=20
        if pack : padx = pady = 5, side = left
        if grid: row = column = 0 sticky = "East"
    """

    def __init__(self, name, validation="", default="", contextualMenu=None, **kwargs):
        """
        Constructor for a form text

        Args:
            name: the entry name (id).
            validation: a regex used to check the input
                             in the checkForm function. default is "" or a callback function
            default: a default value for the Entry, default is ""
            contextualMenu: (Opt.) a contextualMenu to open when right clicked. default is None
            kwargs: same keyword args as you would give to ttk.Text
        """
        super().__init__(name)
        self.validation = validation
        self.default = default
        self.contextualMenu = contextualMenu
        self.kwargs = kwargs
        self.text = None
        self.widgetMenuOpen = None
        self.just_editor = self.getKw("just_editor", False)
        self.enable_preview = self.getKw("enable_preview", True)
        self.style_change = self.getKw("style_change", False)
        self.allow_maximize = self.getKw("allow_maximize", True)

    def close(self):
        """Option of the contextual menu : Close the contextual menu by doing nothing
        """
        pass

    def sanitize(self, string):
        """remove unwanted things in text"""
        return string.replace("\r","")

    def constructView(self, parent):
        """
        Create the text view inside the parent view given

        Args:
            parent: parent FormPanel.
        """
        state = self.getKw("state", "normal")
        dark_mode = self.getKw("dark_mode", False)
        self.mdFrame = TkintermdFrame(parent.panel, default_text=self.default, just_editor=self.just_editor, style_change=self.style_change,
                                       enable_preview=self.enable_preview, height=self.getKw("height", 0), binds=self.getKw("binds", {}),
                                       enable_maximize=self.allow_maximize)
        from pollenisatorgui.core.components.settings import Settings
        s = Settings()
        if dark_mode:
            editor_style = s.local_settings.get("editor_dark_theme", "material")
        else:
            editor_style = s.local_settings.get("editor_light_style","stata-light")
        self.mdFrame.load_style(editor_style)
        if state == "disabled":
            self.mdFrame.text_area.configure(state="disabled")
        if parent.gridLayout:
            self.mdFrame.grid(row=self.getKw("row", 0), column=self.getKw(
                "column", 0), sticky=self.getKw("sticky", tk.EW), **self.kwargs)
        else:
            self.mdFrame.pack(side=self.getKw("side", "left"), padx=self.getKw(
                "padx", 10), pady=self.getKw("pady", 5), expand=self.getKw("expand", True), fill=self.getKw("fill", tk.NONE), **self.kwargs)

    
    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return the entry value as string.
        """
        return self.mdFrame.text_area.get('1.0', 'end-1c')

    def setValue(self, newval):
        """
        Set the text value.
        Args:
            newval: the new value to be set inside the text
        """
        state = self.mdFrame.text_area.cget("state")
        self.mdFrame.text_area.configure(state="normal")
        self.mdFrame.text_area.delete("1.0", "end")
        self.mdFrame.text_area.insert("1.0", self.sanitize(newval))
        self.mdFrame.text_area.configure(state=state)

    def checkForm(self):
        """
        Check if this form is correctly filled.
        Check with the validation given in constructor.

        Returns:
            {
                "correct": True if the form is correctly filled, False otherwise.
                "msg": A message indicating what is not correctly filled.
            }
        """
        if isinstance(self.validation, str):
            import re
            if re.match(self.validation, self.getValue(), re.MULTILINE) is None:
                return False, self.name+" value is incorrect."
            return True, ""
        elif callable(self.validation):
            return self.validation(self.getValue()), self.name+" value is incorrect."


    def setFocus(self):
        """Set the focus to the ttk entry widget.
        """
        self.mdFrame.text_area.focus_set()
